import re
from dataclasses import dataclass
from typing import List, Optional

from nidaqmx.errors import DaqError
from nidaqmx.grpc_session_options import GrpcSessionOptions
from nidaqmx._base_interpreter import BaseInterpreter


# Method logic adapted from
# //Measurements/Infrastructure/dmxf/trunk/2.5/source/nimuck/parseUtilities.cpp

_invalid_range_syntax_message = (
    "Syntax for a range of objects in the input string is invalid.\n\n"
    "For ranges of objects, specify a number immediately before and after "
    "every colon (':') in the input string. Or, if a name is specified after "
    "the colon, it must be identical to the name specified immediately before "
    "the colon. Colons are not allowed within the names of the individual "
    "objects.")


@dataclass
class _ChannelInfo:
    base_name: str = ""
    start_index: int = -1
    start_index_str: str = ""
    end_index: int = -1
    end_index_str: str = ""

    def to_flattened_name(self) -> str:
        """Convert the channel info to a flattened channel name."""
        if self.start_index == -1:
            return self.base_name
        elif self.start_index == self.end_index:
            return f"{self.base_name}{self.start_index_str}"
        else:
            return f"{self.base_name}{self.start_index_str}:{self.end_index_str}"


def flatten_channel_string(channel_names: List[str]) -> str:
    """
    Converts a list of channel names to a comma-delimited list of names.

    You can use this method to convert a list of physical or virtual channel
    names to a single string prior to using the DAQmx Create Channel methods or
    instantiating a DAQmx Task object.

    Note: For simplicity, this implementation is not fully compatible with the
    NI-DAQmx driver implementation, which is generally more permissive. For
    example, the driver is more graceful with whitespace padding. It was deemed
    valuable to implement this natively in Python, so it can be leveraged in
    workflows that don't have the driver installed. If we have specific examples
    where this approximation is a problem, we can revisit this in the future.

    Args:
        channel_names: The list of physical or virtual channel names.
    Returns:
        The resulting comma-delimited list of physical or virtual channel names.
    """
    unflattened_channel_names = []
    for channel_name in channel_names:
        unflattened_channel_names.extend(unflatten_channel_string(channel_name))

    # Go through the channel names and flatten them.
    flattened_channel_list = []
    previous = _ChannelInfo()
    for channel_name in unflattened_channel_names:
        m = re.search('(.*[^0-9])?([0-9]+)$', channel_name)
        if not m:
            # If the channel name doesn't end in a valid number, just use the
            # channel name as-is.
            flattened_channel_list.append(previous.to_flattened_name())
            previous = _ChannelInfo(channel_name)
        else:
            # If the channel name ends in a valid number, we may need to flatten
            # this channel with subsequent channels in the x:y format.
            current_base_name = m.group(1)
            current_index_str = m.group(2)
            current_index = int(current_index_str)

            if current_base_name == previous.base_name and (
                (current_index == previous.end_index + 1 and
                 previous.end_index >= previous.start_index) or
                (current_index == previous.end_index - 1 and
                 previous.end_index <= previous.start_index)):
                # If the current channel name has the same base name as the
                # previous and it's end index differs by 1, change the end
                # index value. It gets flattened later.
                previous.end_index = current_index
                previous.end_index_str = current_index_str
            else:
                # If the current channel name has the same base name as the
                # previous or it's end index differs by more than 1, it doesn't
                # get flattened with the previous channel.
                flattened_channel_list.append(previous.to_flattened_name())
                previous = _ChannelInfo(
                    current_base_name, 
                    current_index, 
                    current_index_str, 
                    current_index, 
                    current_index_str
                )

    # Convert the final channel dictionary to a flattened string
    flattened_channel_list.append(previous.to_flattened_name())

    # Remove empty strings in list, convert to comma-delimited string, then trim
    # whitespace.
    return ','.join([_f for _f in flattened_channel_list if _f]).strip()

                                   
def unflatten_channel_string(channel_names: str) -> List[str]:
    """
    Converts a comma-delimited list of channel names to a list of names.

    You can use this method to convert a comma-delimited list or range of
    physical or virtual channels into a list of physical or virtual channel
    names.

    Note: For simplicity, this implementation is not fully compatible with the
    NI-DAQmx driver implementation, which is generally more permissive. For
    example, the driver is more graceful with whitespace padding. It was deemed
    valuable to implement this natively in Python, so it can be leveraged in
    workflows that don't have the driver installed. If we have specific examples
    where this approximation is a problem, we can revisit this in the future.

    Args:
        channel_names: The list or range of physical or virtual channels.
        
    Returns:
        The list of physical or virtual channel names. 
        
        Each element of the list contains a single channel.
    """
    channel_list_to_return = []
    channel_list = [c for c in channel_names.strip().split(',') if c]

    for channel in channel_list:
        channel = channel.strip()
        colon_index = channel.find(':')

        if colon_index == -1:
            channel_list_to_return.append(channel)
        else:
            before = channel[:colon_index]
            after = channel[colon_index+1:]

            m_before = re.match('(.*?)([0-9]+)$', before)
            m_after = re.match('(.*?)([0-9]+)$', after)

            if not m_before or not m_after:
                raise DaqError(_invalid_range_syntax_message,
                                       error_code=-200498)

            if m_after.group(1) and (
                m_before.group(1).lower() != m_after.group(1).lower()):
                raise DaqError(_invalid_range_syntax_message,
                               error_code=-200498)

            num_before_str = m_before.group(2)
            num_before = int(num_before_str)
            num_after_str = m_after.group(2)
            num_after = int(num_after_str)

            num_min_width = 0
            # If there are any leading 0s in the first number, we want to ensure
            # match that width. This is established precedence in the DAQmx
            # algorithm.
            if num_before > 0 and len(num_before_str.lstrip('0')) < len(num_before_str):
                num_min_width = len(num_before_str)

            num_max = max([num_before, num_after])
            num_min = min([num_before, num_after])
            number_of_channels = (num_max - num_min) + 1

            if number_of_channels >= 15000:
                raise DaqError(_invalid_range_syntax_message,
                                       error_code=-200498)

            colon_expanded_channel = []
            for i in range(number_of_channels):
                current_number = num_min + i
                if num_min_width > 0:
                    # Using fstrings to create format strings. Braces for days!
                    zero_padded_format_specifier = f"{{:0{num_min_width}d}}"
                    current_number_str = zero_padded_format_specifier.format(current_number)
                else:
                    current_number_str = str(current_number)
                colon_expanded_channel.append(f"{m_before.group(1)}{current_number_str}")

            if num_after < num_before:
                colon_expanded_channel.reverse()

            channel_list_to_return.extend(colon_expanded_channel)

    return channel_list_to_return


def _select_interpreter(
    grpc_options: Optional[GrpcSessionOptions] = None,
    interpreter: Optional[BaseInterpreter] = None
) -> BaseInterpreter:
    if interpreter:
        return interpreter
    else:
        if grpc_options:
            from nidaqmx._grpc_interpreter import GrpcStubInterpreter
            return GrpcStubInterpreter(grpc_options)
        else:
            from nidaqmx._library_interpreter import LibraryInterpreter
            return LibraryInterpreter()



from collections.abc import Sequence

from nidaqmx.errors import DaqError
from nidaqmx.error_codes import DAQmxErrors
from nidaqmx.system.physical_channel import PhysicalChannel, _PhysicalChannelAlternateConstructor
from nidaqmx.utils import unflatten_channel_string, flatten_channel_string

class PhysicalChannelCollection(Sequence):
    """
    Contains the collection of physical channels for a DAQmx device.
    
    This class defines methods that implements a container object.
    """
    def __init__(self, device_name, interpreter):
        """
        Do not construct this object directly; instead, construct a nidaqmx.system.Device and use the appropriate property, such as device.ai_physical_channels.
        """
        self._name = device_name
        self._interpreter = interpreter

    def __contains__(self, item):
        channel_names = self.channel_names

        if isinstance(item, str):
            items = unflatten_channel_string(item)
            return all([i in channel_names for i in items])
        elif isinstance(item, PhysicalChannel):
            return item._name in channel_names
        return False

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self._name == other._name
        return False

    def __getitem__(self, index):
        """
        Indexes a subset of physical channels on this physical channel
        collection.

        Args:
            index: The value of the index. The following index types
                are supported:
                - str: Name of the physical channel, without the
                    device name prefix, e.g. 'ai0'. You also can
                    specify a string that contains a list or range of
                    names to this input. If you have a list of names,
                    use the DAQmx Flatten Channel String function to
                    convert the list to a string.
                - int: Index/position of the physical channel in the
                    collection.
                - slice: Range of the indexes/positions of physical
                    channels in the collection.
        Returns:
            nidaqmx.system.physical_channel.PhysicalChannel: 
            
            Indicates the subset of physical channels indexed.
        """
        if isinstance(index, int):
            return _PhysicalChannelAlternateConstructor(self.channel_names[index], self._interpreter)
        elif isinstance(index, slice):
            return [_PhysicalChannelAlternateConstructor(channel, self._interpreter) for channel in self.channel_names[index]]
        elif isinstance(index, str):
            requested_channels = unflatten_channel_string(index)
            # Ensure the channel names are fully qualified. If the channel is invalid, the user will get errors from the
            # channel objects on use.
            channels_to_use = []
            for channel in requested_channels:
                if channel.startswith(f"{self._name}/"):
                    channels_to_use.append(channel)
                else:
                    channels_to_use.append(f'{self._name}/{channel}')

            if len(channels_to_use) == 1:
                return _PhysicalChannelAlternateConstructor(channels_to_use[0], self._interpreter)
            return [_PhysicalChannelAlternateConstructor(channel, self._interpreter) for channel in channels_to_use]
        else:
            raise DaqError(
                'Invalid index type "{}" used to access collection.'
                .format(type(index)), DAQmxErrors.UNKNOWN)

    def __iter__(self):
        for channel_name in self.channel_names:
            yield _PhysicalChannelAlternateConstructor(channel_name, self._interpreter)

    def __len__(self):
        return len(self.channel_names)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __reversed__(self):
        channel_names = self.channel_names
        channel_names.reverse()

        for channel_name in channel_names:
            yield _PhysicalChannelAlternateConstructor(channel_name, self._interpreter)

    @property
    def all(self):
        """
        nidaqmx.system.physical_channel.PhysicalChannel: Specifies a
            physical channel object that represents the entire list of
            physical channels on this channel collection.
        """
        return _PhysicalChannelAlternateConstructor(flatten_channel_string(self.channel_names), self._interpreter)

    @property
    def channel_names(self):
        """
        List[str]: Specifies the entire list of physical channels in this
            collection.
        """
        raise NotImplementedError()


class AIPhysicalChannelCollection(PhysicalChannelCollection):
    """
    Contains the collection of analog input physical channels for a
    DAQmx device.
    
    This class defines methods that implements a container object.
    """

    @property
    def channel_names(self):
        val = self._interpreter.get_device_attribute_string(self._name, 0x231e)
        return unflatten_channel_string(val)


class AOPhysicalChannelCollection(PhysicalChannelCollection):
    """
    Contains the collection of analog output physical channels for a
    DAQmx device.
    
    This class defines methods that implements a container object.
    """

    @property
    def channel_names(self):
        val = self._interpreter.get_device_attribute_string(self._name, 0x231f)
        return unflatten_channel_string(val)


class CIPhysicalChannelCollection(PhysicalChannelCollection):
    """
    Contains the collection of counter input physical channels for a
    DAQmx device.
    
    This class defines methods that implements a container object.
    """

    @property
    def channel_names(self):
        val = self._interpreter.get_device_attribute_string(self._name, 0x2324)
        return unflatten_channel_string(val)


class COPhysicalChannelCollection(PhysicalChannelCollection):
    """
    Contains the collection of counter output physical channels for a
    DAQmx device.
    
    This class defines methods that implements a container object.
    """

    @property
    def channel_names(self):
        val = self._interpreter.get_device_attribute_string(self._name, 0x2325)
        return unflatten_channel_string(val)


class DILinesCollection(PhysicalChannelCollection):
    """
    Contains the collection of digital input lines for a DAQmx device.
    
    This class defines methods that implements a container object.
    """

    @property
    def channel_names(self):
        val = self._interpreter.get_device_attribute_string(self._name, 0x2320)
        return unflatten_channel_string(val)


class DOLinesCollection(PhysicalChannelCollection):
    """
    Contains the collection of digital output lines for a DAQmx device.
    
    This class defines methods that implements a container object.
    """

    @property
    def channel_names(self):
        val = self._interpreter.get_device_attribute_string(self._name, 0x2322)
        return unflatten_channel_string(val)


class DIPortsCollection(PhysicalChannelCollection):
    """
    Contains the collection of digital input ports for a DAQmx device.
    
    This class defines methods that implements a container object.
    """

    @property
    def channel_names(self):
        val = self._interpreter.get_device_attribute_string(self._name, 0x2321)
        return unflatten_channel_string(val)


class DOPortsCollection(PhysicalChannelCollection):
    """
    Contains the collection of digital output ports for a DAQmx device.
    
    This class defines methods that implements a container object.
    """

    @property
    def channel_names(self):
        val = self._interpreter.get_device_attribute_string(self._name, 0x2323)
        return unflatten_channel_string(val)



from collections.abc import Sequence

from nidaqmx.task.channels._channel import Channel
from nidaqmx.errors import DaqError
from nidaqmx.error_codes import DAQmxErrors
from nidaqmx.utils import unflatten_channel_string, flatten_channel_string


class ChannelCollection(Sequence):
    """
    Contains the collection of channels for a DAQmx Task.
    
    This class defines methods that implements a container object.
    """
    def __init__(self, task_handle, interpreter):
        """
        Do not construct this object directly; instead, construct a nidaqmx.Task and use the appropriate property, such as task.ai_channels.
        """
        self._handle = task_handle
        self._interpreter = interpreter

    def __contains__(self, item):
        channel_names = self.channel_names

        if isinstance(item, str):
            items = unflatten_channel_string(item)
        elif isinstance(item, Channel):
            items = item.channel_names

        return all([item in channel_names for item in items])

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self._handle == other._handle
        return False

    def __getitem__(self, index):
        """
        Indexes a subset of virtual channels on this channel collection.

        Args:
            index: The value of the index. The following index types are
                supported:
                - str: Name of the virtual channel. You also can specify a
                    string that contains a list or range of names to this
                    input. If you have a list of names, use the DAQmx
                    Flatten Channel String function to convert the list to a
                    string.
                - int: Index/position of the virtual channel in the collection.
                - slice: Range of the indexes/positions of virtual channels in
                    the collection.
        Returns:
            nidaqmx.task.channels.Channel: 
            
            Indicates a channel object representing the subset of virtual
            channels indexed.
        """
        if isinstance(index, int):
            channel_names = self.channel_names[index]
        elif isinstance(index, slice):
            channel_names = flatten_channel_string(self.channel_names[index])
        elif isinstance(index, str):
            channel_names = index
        else:
            raise DaqError(
                'Invalid index type "{}" used to access channels.'
                .format(type(index)), DAQmxErrors.UNKNOWN)

        if channel_names:
            return Channel._factory(self._handle, channel_names, self._interpreter)
        else:
            raise DaqError(
                'You cannot specify an empty index when indexing channels.\n'
                'Index used: {}'.format(index), DAQmxErrors.UNKNOWN)

    def __hash__(self):
        return self._interpreter.hash_task_handle(self._handle)

    def __iter__(self):
        for channel_name in self.channel_names:
            yield Channel._factory(self._handle, channel_name, self._interpreter)

    def __len__(self):
        return len(self.channel_names)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __reversed__(self):
        channel_names = self.channel_names
        channel_names.reverse()

        for channel_name in channel_names:
            yield Channel._factory(self._handle, channel_name, self._interpreter)

    @property
    def all(self):
        """
        :class:`nidaqmx.task.channels.Channel`:
            Specifies a channel object that represents the entire list of 
            virtual channels on this channel collection.
        """
        # Passing a blank string means all channels.
        return Channel._factory(self._handle, '', self._interpreter)

    @property
    def channel_names(self):
        """
        List[str]: Specifies the entire list of virtual channels on this
            channel collection.
        """
        val = self._interpreter.get_task_attribute_string(self._handle, 0x1273)
        return unflatten_channel_string(val)
